"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProxyServer = void 0;
const http_1 = require("http");
const https_1 = require("https");
const tls_1 = require("tls");
const net = __importStar(require("net"));
const koa_compose_1 = __importDefault(require("koa-compose"));
const proxy_pass_middleware_1 = require("./middlewares/proxy-pass-middleware");
const certification_manager_1 = require("./certification-manager");
const const_1 = require("./const");
const url_middleware_1 = require("./middlewares/url-middleware");
const proxy_rule_middleware_1 = require("./middlewares/proxy-rule-middleware");
const rule_manager_1 = require("./rule-manager");
const upgrade_middleware_1 = require("./middlewares/upgrade-middleware");
const utils_1 = require("./utils");
class ProxyServer {
    proxyServer = new http_1.Server();
    httpsServer = new https_1.Server();
    ruleManager = new rule_manager_1.RuleManager();
    certificationManager = new certification_manager_1.CertificationManager();
    /**
     * 初始化必要的服务
     *
     * @author yuzhanglong
     * @date 2021-08-11 00:48:00
     */
    async initServers() {
        // 证书和 server 初始化
        const httpsServerCert = await this.certificationManager.createCertificationByDomain('localhost');
        this.proxyServer = new http_1.Server();
        this.httpsServer = new https_1.Server({
            cert: httpsServerCert.cert,
            key: httpsServerCert.key,
            SNICallback: async (servername, callback) => {
                const { key, cert } = await this.certificationManager.createCertificationByDomain(servername);
                callback(null, tls_1.createSecureContext({
                    key,
                    cert,
                }));
            },
        });
        // compose 中间件
        const handlers = koa_compose_1.default([
            url_middleware_1.createUrlMiddleWare(),
            proxy_rule_middleware_1.createProxyRuleMiddleware(this.ruleManager),
            upgrade_middleware_1.createUpgradeMiddleware(),
            proxy_pass_middleware_1.createProxyPassMiddleware(),
        ]);
        // 在收到 XHR 的代理请求时做些什么
        this.proxyServer.on('request', async (req, res) => {
            await handlers({
                incomingRequestData: req,
                proxyServerResponse: res,
                protocol: 'http',
            });
        });
        // 在收到通过 CONNECT 方法建立而来的隧道传来的数据后做些什么
        this.httpsServer.on('request', async (req, res) => {
            await handlers({
                incomingRequestData: req,
                proxyServerResponse: res,
                protocol: 'https',
            });
        });
        // 在收到 HTTP CONNECT 请求时做些什么
        // 如果我们要代理的请求为 https 协议，如果将请求代理到本服务器，浏览器会发送一个 CONNECT 请求，建立双向隧道
        // 在 nodejs 的官方文档中提供了这样一个 DEMO：http://nodejs.cn/api/http.html#http_event_connect
        // 于是我们可以将这个请求转发到一个新的服务器上，然后让这个新的服务器去请求实际资源
        this.proxyServer.on('connect', (req, clientSocket, head) => {
            // 只有 https 请求走代理才会走到这一步，所以直接连接我们的 https 服务器就可以了
            const { domain, port } = utils_1.divideConnectMethodReqUrl(req.url);
            // 是否匹配用户配置的域名, 如果不匹配我们不用走 proxy 服务器，这样一是可以优化性能
            // 二是防止某些服务器使用了 TLS Pinning 技术，会对证书的摘要进行验证
            const isProxyMatched = this.ruleManager.matchDomain(domain);
            const finalDomain = isProxyMatched ? const_1.LOCAL_HOST : domain;
            const finalPort = isProxyMatched ? const_1.PROXY_PASS_SERVICE_PORT : port;
            const proxyPassServiceSocket = net.connect(finalPort, finalDomain, () => {
                clientSocket.write(`HTTP/${req.httpVersion} 200 OK\r\n\r\n`, 'utf-8', () => {
                    // 隧道流的第一个数据包（可能为空）
                    proxyPassServiceSocket.write(head);
                    proxyPassServiceSocket.pipe(clientSocket);
                    clientSocket.pipe(proxyPassServiceSocket);
                });
                proxyPassServiceSocket.on('error', (e) => {
                    console.log('error!');
                    console.log(e);
                });
            });
            clientSocket.on('error', (e) => {
                console.log(e);
            });
        });
        // 在收到 http 升级请求时做些什么
        this.proxyServer.on('upgrade', async (req, socket, head) => {
            await handlers({
                incomingRequestData: req,
                socketBetweenClientAndProxyServer: socket,
                head,
                protocol: 'ws',
            });
        });
        // 在收到 https 升级请求时做些什么
        this.httpsServer.on('upgrade', async (req, socket, head) => {
            await handlers({
                incomingRequestData: req,
                socketBetweenClientAndProxyServer: socket,
                head,
                protocol: 'wss',
            });
        });
    }
    /**
     * 监听端口
     *
     * @author yuzhanglong
     * @date 2021-08-11 13:07:59
     */
    async listen() {
        if (!this.httpsServer || !this.proxyServer)
            throw new Error('[@attachments/proxy] please init server!');
        this.proxyServer.listen(const_1.PROXY_SERVER_PORT, const_1.LOCAL_HOST, () => {
            console.log('[@attachments/proxy] proxy server is running...');
        });
        this.httpsServer.listen(const_1.PROXY_PASS_SERVICE_PORT, const_1.LOCAL_HOST, () => {
            console.log('[@attachments/proxy] https server is running...');
        });
    }
    /**
     * 增加代理规则
     *
     * @author yuzhanglong
     * @date 2021-08-09 00:11:39
     * @param domain
     * @param ruleConfig
     */
    addRule(domain, ...ruleConfig) {
        this.ruleManager.addRule(domain, ...ruleConfig);
    }
    /**
     * 清除所有的代理规则
     *
     * @author yuzhanglong
     * @date 2021-12-16 00:12:46
     */
    clearRules() {
        this.ruleManager.clear();
    }
}
exports.ProxyServer = ProxyServer;
//# sourceMappingURL=proxy-server.js.map